
import pandas as pd
import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import RandomizedSearchCV
from .config import Config

class ModelTrainer:
    def __init__(self, random_state=None):
        self.random_state = random_state if random_state else Config.RANDOM_STATE
        self.model = None
        self.best_params = None
        self.cv_results = None
        
    def train(self, X, y, param_distributions=None, n_iter=10, cv=5, scoring='roc_auc'):
        """
        Train Logistic Regression with Randomized Search.
        """
        if param_distributions is None:
            param_distributions = {
                'C': np.logspace(-4, 4, 20),
                'penalty': ['l2'], # 'l1', 'elasticnet' require solver tuning, keeping simple
                'solver': ['lbfgs', 'liblinear']
            }
            
        lr = LogisticRegression(random_state=self.random_state, max_iter=1000)
        
        search = RandomizedSearchCV(
            estimator=lr,
            param_distributions=param_distributions,
            n_iter=n_iter,
            cv=cv,
            scoring=scoring,
            random_state=self.random_state,
            n_jobs=1, # Avoid joblib parallel issues on Windows or small data
            verbose=0
        )
        
        search.fit(X, y)
        
        self.model = search.best_estimator_
        self.best_params = search.best_params_
        self.cv_results = pd.DataFrame(search.cv_results_)
        
        return self.model
        
    def set_coefficients(self, coefs, intercept):
        """
        Manually set coefficients for the model.
        coefs: array-like of shape (n_features,)
        intercept: float
        """
        if self.model is None:
            # Initialize a dummy model if not trained
            self.model = LogisticRegression(random_state=self.random_state)
            # Need to initialize classes_ etc. for predict_proba to work
            # This is tricky without fitting.
            # Best practice: Fit on dummy data or ensure train() is called first.
            raise ValueError("Model must be trained/initialized first.")
            
        if hasattr(self.model, 'best_estimator_'):
             estimator = self.model.best_estimator_
        else:
             estimator = self.model
             
        estimator.coef_ = np.array([coefs])
        estimator.intercept_ = np.array([intercept])

    def predict_proba(self, X):
        if self.model is None:
            raise ValueError("Model not trained.")
        return self.model.predict_proba(X)[:, 1]




import unittest
import pandas as pd
import numpy as np
import os
from autoscore import AutoScore, BinningProcess

class TestAutoScore(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        # Create synthetic data
        np.random.seed(42)
        n = 500
        cls.df = pd.DataFrame({
            'age': np.random.randint(18, 70, n),
            'income': np.random.normal(50000, 15000, n),
            'target': np.random.binomial(1, 0.3, n),
            'category': np.random.choice(['A', 'B', 'C'], n)
        })
        
    def test_binning_process(self):
        binning = BinningProcess(max_bins=5)
        binning.fit(self.df[['age', 'income', 'category']], self.df['target'])
        
        self.assertIn('age', binning.rules)
        self.assertIn('income', binning.rules)
        self.assertIn('category', binning.rules)
        
        # Check transformation
        X_woe = binning.transform(self.df[['age', 'income', 'category']])
        self.assertEqual(X_woe.shape, (500, 3))
        self.assertFalse(X_woe.isnull().any().any())

    def test_full_pipeline(self):
        score = AutoScore(random_state=42)
        score.fit(
            X=self.df.drop(columns=['target']),
            y=self.df['target'],
            max_bins=5,
            selection_method='stepwise'
        )
        
        self.assertIsNotNone(score.scorecard_df)
        self.assertIsNotNone(score.model_metrics)
        self.assertIn('train', score.model_metrics)
        
        # Test Prediction
        probs = score.predict_proba(self.df.drop(columns=['target']))
        self.assertEqual(len(probs), 500)
        
    def test_reporting(self):
        score = AutoScore(random_state=42)
        score.fit(self.df.drop(columns=['target']), self.df['target'])
        
        report_file = 'test_report.xlsx'
        score.create_report(report_file)
        self.assertTrue(os.path.exists(report_file))
        os.remove(report_file)

if __name__ == '__main__':
    unittest.main()




import pandas as pd
import numpy as np

class FeatureAnalyzer:
    def __init__(self):
        self.summary = None
        self.drop_cols = []
        
    def analyze(self, df, target=None):
        """
        Analyze dataset and return summary statistics.
        """
        stats = []
        for col in df.columns:
            series = df[col]
            dtype = series.dtype
            
            # Count unique values
            nunique = series.nunique()
            
            # Missing rate
            missing = series.isnull().sum()
            missing_rate = missing / len(df)
            
            # Basic stats
            if np.issubdtype(dtype, np.number):
                min_val = series.min()
                max_val = series.max()
                mean_val = series.mean()
                std_val = series.std()
                zeros = (series == 0).sum()
                zeros_rate = zeros / len(df)
            else:
                min_val = np.nan
                max_val = np.nan
                mean_val = np.nan
                std_val = np.nan
                zeros = 0
                zeros_rate = 0
                
            stats.append({
                'variable': col,
                'dtype': str(dtype),
                'nunique': nunique,
                'missing_rate': missing_rate,
                'min': min_val,
                'max': max_val,
                'mean': mean_val,
                'std': std_val,
                'zeros_rate': zeros_rate
            })
            
        self.summary = pd.DataFrame(stats)
        return self.summary
        
    def get_drop_columns(self, threshold_missing=0.95, threshold_unique=1):
        """
        Identify columns to drop based on missing rate or single value.
        """
        if self.summary is None:
            raise ValueError("Run analyze() first.")
            
        drop_missing = self.summary[self.summary['missing_rate'] > threshold_missing]['variable'].tolist()
        drop_unique = self.summary[self.summary['nunique'] <= threshold_unique]['variable'].tolist()
        
        self.drop_cols = list(set(drop_missing + drop_unique))
        return self.drop_cols




# AutoScore - Automated Scorecard Modeling Library

from .pipeline import AutoScore
from .eda import FeatureAnalyzer
from .binning import BinningProcess
from .selection import FeatureSelector
from .modeling import ModelTrainer
from .evaluation import ModelEvaluator
from .scorecard import ScorecardTransformer
from .reporting import ReportGenerator

__version__ = '0.1.0'




import pandas as pd
import numpy as np
from sklearn.tree import DecisionTreeClassifier
from .config import Config

class BinningProcess:
    def __init__(self, method='chi', max_bins=10, min_bin_pct=0.05, special_values=None):
        self.method = method
        self.max_bins = max_bins
        self.min_bin_pct = min_bin_pct
        self.special_values = special_values if special_values else Config.SPECIAL_VALUES
        self.rules = {}  # Stores binning rules: {col: {'cutoffs': [], 'woe': {}}}
        self.iv_df = None
        

    def set_rules(self, rules):
        """
        Manually set binning rules.
        """
        self.rules = rules

    def recalculate_stats(self, X, y):
        """
        Recalculate WOE and IV for existing rules on new data.
        """
        iv_stats = []
        for col, rule in self.rules.items():
            if col not in X.columns:
                continue
                
            if rule['type'] == 'numeric':
                cutoffs = rule['cutoffs']
                woe_map, iv, bin_stats = self._calculate_woe(X[col], y, cutoffs)
            elif rule['type'] == 'categorical':
                categories = rule['categories']
                woe_map, iv, bin_stats = self._calculate_woe_categorical(X[col], y, categories)
                
            # Update rule with new stats (keep cutoffs/categories)
            self.rules[col]['woe_map'] = woe_map
            self.rules[col]['iv'] = iv
            self.rules[col]['stats'] = bin_stats
            
            iv_stats.append({'variable': col, 'iv': iv})
            
        self.iv_df = pd.DataFrame(iv_stats).sort_values('iv', ascending=False)

    def fit(self, X, y, monotonicity_constraints=None):
        """
        Fit binning rules for each column in X using y.
        monotonicity_constraints: dict {col: 'increasing'|'decreasing'|'u-shape'}
        """
        self.rules = {}
        iv_stats = []
        
        for col in X.columns:
            if X[col].dtype == 'object' or X[col].dtype.name == 'category':
                # Categorical binning
                categories = self._categorical_binning(X[col], y, max_bins=self.max_bins)
                woe_map, iv, bin_stats = self._calculate_woe_categorical(X[col], y, categories)
                
                self.rules[col] = {
                    'categories': categories, # List of lists (bins)
                    'woe_map': woe_map,
                    'iv': iv,
                    'stats': bin_stats,
                    'type': 'categorical'
                }
            else:
                # Numerical binning
                constraint = monotonicity_constraints.get(col) if monotonicity_constraints else None
                
                if self.method == 'chi':
                    cutoffs = self._chimerge(X[col], y, max_bins=self.max_bins, constraint=constraint)
                elif self.method == 'dt':
                    cutoffs = self._dt_binning(X[col], y, max_bins=self.max_bins)
                else:
                    cutoffs = self._quantile_binning(X[col], max_bins=self.max_bins)
                    
                # Calculate WOE and IV
                woe_map, iv, bin_stats = self._calculate_woe(X[col], y, cutoffs)
                
                self.rules[col] = {
                    'cutoffs': cutoffs,
                    'woe_map': woe_map,
                    'iv': iv,
                    'stats': bin_stats,
                    'type': 'numeric'
                }
            
            iv_stats.append({'variable': col, 'iv': iv})
            
        self.iv_df = pd.DataFrame(iv_stats).sort_values('iv', ascending=False)
        return self

    def transform(self, X):
        """
        Apply WOE transformation.
        """
        X_woe = X.copy()
        for col, rule in self.rules.items():
            if col not in X.columns:
                continue
            
            if rule['type'] == 'numeric':
                cutoffs = rule['cutoffs']
                woe_map = rule['woe_map']
                
                # Binning
                bins = [-np.inf] + cutoffs + [np.inf]
                
                X_woe[col] = pd.cut(X[col], bins=bins, labels=False, include_lowest=True)
                X_woe[col] = X_woe[col].map(woe_map).astype(float)
                
                if -1 in woe_map: # Convention: -1 for NaN
                    X_woe[col] = X_woe[col].fillna(woe_map[-1])
                    
            elif rule['type'] == 'categorical':
                woe_map = rule['woe_map']
                # woe_map for categorical is {value: woe}
                # But we stored it by bin index in _calculate_woe_categorical
                # So we need to map value -> bin_index -> woe
                
                # Create a direct map: value -> woe
                value_woe_map = {}
                categories = rule['categories'] # List of lists
                
                for bin_idx, bin_values in enumerate(categories):
                    woe = woe_map[bin_idx]
                    for val in bin_values:
                        value_woe_map[val] = woe
                        
                # Map values
                X_woe[col] = X[col].map(value_woe_map)
                
                # Handle unknown values (treat as missing or new bin?)
                # For now, map to NaN/missing bin if exists
                if -1 in woe_map:
                    X_woe[col] = X_woe[col].fillna(woe_map[-1])
                else:
                    # If no missing bin, fill with 0 or mean woe? 
                    # Let's fill with 0 (neutral) for safety or error?
                    X_woe[col] = X_woe[col].fillna(0) 
                    
                X_woe[col] = X_woe[col].astype(float)

        return X_woe

    def _categorical_binning(self, x, y, max_bins=10):
        """
        Bin categorical features by grouping similar bad rates.
        """
        df = pd.DataFrame({'x': x, 'y': y})
        # Fill NA with 'Missing' for analysis
        df['x'] = df['x'].fillna('Missing')
        
        # Calculate Bad Rate per category
        grouped = df.groupby('x', observed=False)['y'].agg(['count', 'sum'])
        grouped['bad_rate'] = grouped['sum'] / grouped['count']
        grouped = grouped.sort_values('bad_rate')
        
        # We can treat bad_rate as a continuous variable and use ChiMerge?
        # Or just use the ordered categories as "x" for ChiMerge.
        
        # Strategy: 
        # 1. Sort categories by Bad Rate.
        # 2. Treat them as ordinal (0, 1, 2...).
        # 3. Apply ChiMerge on these ordinal values.
        
        categories_ordered = grouped.index.tolist()
        cat_map = {cat: i for i, cat in enumerate(categories_ordered)}
        
        df['x_ordinal'] = df['x'].map(cat_map)
        
        # Use ChiMerge on x_ordinal
        cutoffs = self._chimerge(df['x_ordinal'], df['y'], max_bins=max_bins)
        
        # Convert cutoffs back to groups of categories
        # cutoffs are indices in categories_ordered
        # e.g., cutoffs = [2, 5] means bins: [0,1,2], [3,4,5], [6...]
        
        bins = []
        start_idx = 0
        
        # Ensure cutoffs are integers
        cutoffs = [int(c) for c in cutoffs]
        
        for end_idx in cutoffs:
            # bin includes end_idx (because chimerge cutoffs are inclusive on left in my logic? 
            # Wait, _chimerge returns right edges of intervals from pd.cut(include_lowest=True).
            # pd.cut: (a, b]. 
            # My _chimerge uses pd.cut then aggregates. 
            # Let's assume cutoffs are the split points.
            # If cutoffs=[2], splits are <=2 and >2.
            
            # The indices are 0 to N-1.
            # If cutoff is 2.5, it splits 0,1,2 | 3,4...
            
            # _chimerge returns specific values from x as cutoffs.
            # Since x are integers 0..N, cutoffs will be integers (or floats if mean).
            # In my _chimerge, I take right edge of interval.
            # If unique values, cutoffs are values.
            
            # Let's look at _chimerge again.
            # "cutoffs = grouped.index.tolist()[:-1]" (for small unique)
            # So cutoffs are the values.
            # If values are 0,1,2,3. Cutoffs could be 0, 1, 2.
            # If cutoff=0, split is <=0.
            
            end_idx = int(end_idx)
            current_bin = categories_ordered[start_idx : end_idx + 1]
            bins.append(current_bin)
            start_idx = end_idx + 1
            
        # Add last bin
        if start_idx < len(categories_ordered):
            bins.append(categories_ordered[start_idx:])
            
        return bins

    def _calculate_woe_categorical(self, x, y, categories):
        """
        Calculate WOE for categorical bins.
        """
        # categories is list of lists
        # Map x to bin index
        val_to_bin = {}
        for idx, cat_list in enumerate(categories):
            for val in cat_list:
                val_to_bin[val] = idx
        
        # Handle Missing if not in categories (though we handled it in binning)
        # If x has real NaNs and we mapped 'Missing' in binning, we need to map NaN here too?
        # x is original series.
        
        df = pd.DataFrame({'x': x, 'y': y})
        # We need to temporarily fillna to map
        df['x_temp'] = df['x'].fillna('Missing')
        
        df['bin'] = df['x_temp'].map(val_to_bin)
        
        # Handle values not in map (e.g. new values not seen in fit) -> NaN bin or -1
        df['bin'] = df['bin'].fillna(-1)
        
        total_good = (df['y'] == 0).sum()
        total_bad = (df['y'] == 1).sum()
        
        grouped = df.groupby('bin', observed=False)['y'].agg(['count', 'sum'])
        grouped['bad'] = grouped['sum']
        grouped['good'] = grouped['count'] - grouped['sum']
        
        woe_map = {}
        iv_total = 0
        bin_stats = []
        
        for bin_idx, row in grouped.iterrows():
            good = row['good']
            bad = row['bad']
            
            if good == 0: good = 0.5
            if bad == 0: bad = 0.5
            
            dist_good = good / total_good
            dist_bad = bad / total_bad
            
            woe = np.log(dist_bad / dist_good)
            iv = (dist_bad - dist_good) * woe
            
            woe_map[bin_idx] = woe
            iv_total += iv
            
            # For categorical, min/max are not relevant, but list of values
            if bin_idx != -1 and bin_idx < len(categories):
                bin_label = str(categories[int(bin_idx)])[:50] + "..." # Truncate for display
            else:
                bin_label = "Unmapped"
                
            bin_stats.append({
                'bin_index': bin_idx,
                'min': bin_label, 
                'max': bin_label,
                'good': row['good'],
                'bad': row['bad'],
                'total': row['count'],
                'bad_rate': row['bad'] / row['count'],
                'woe': woe,
                'iv': iv
            })
            
        return woe_map, iv_total, bin_stats



    def _detect_trend(self, x, y):
        """
        Detect optimal trend (1: increasing, -1: decreasing) using Spearman correlation.
        """
        correlation = pd.concat([x, y], axis=1).corr(method='spearman').iloc[0, 1]
        if abs(correlation) < 0.05: # Weak correlation, treat as no constraint? Or weak constraint.
            # Usually we prefer a direction even if weak.
            return 1 if correlation >= 0 else -1
        return 1 if correlation > 0 else -1

    def _enforce_monotonicity(self, stats, cutoffs, constraint):
        """
        Post-processing to enforce monotonicity.
        stats: [[good, bad], ...]
        cutoffs: list of cutoffs
        constraint: 1 (increasing), -1 (decreasing)
        """
        while len(stats) > 2: # Don't merge if only 2 bins
            bad_rates = stats[:, 1] / stats.sum(axis=1)
            violations = []
            
            for i in range(len(bad_rates) - 1):
                if constraint == 1: # Increasing
                    if bad_rates[i] > bad_rates[i+1]:
                        violations.append(i)
                elif constraint == -1: # Decreasing
                    if bad_rates[i] < bad_rates[i+1]:
                        violations.append(i)
            
            if not violations:
                break
                
            # Merge logic:
            # 1. Prefer merging small bins involved in violation
            # 2. Else merge pair with min Chi2
            
            merge_idx = -1
            min_chi2 = np.inf
            
            for idx in violations:
                # Calculate Chi2 for the violating pair
                chi2 = self._calculate_chi2(stats[idx], stats[idx+1])
                if chi2 < min_chi2:
                    min_chi2 = chi2
                    merge_idx = idx
            
            # Execute Merge
            stats[merge_idx] += stats[merge_idx+1]
            stats = np.delete(stats, merge_idx+1, axis=0)
            del cutoffs[merge_idx]
            
        return stats, cutoffs

    def _chimerge(self, x, y, max_bins=10, constraint=None):
        """
        ChiMerge algorithm implementation.
        """
        df = pd.DataFrame({'x': x, 'y': y})
        df = df.dropna()
        
        # Initial fine binning (e.g. 50 bins or unique values if small)
        if df['x'].nunique() > 50:
            _, bins = pd.qcut(df['x'], 50, retbins=True, duplicates='drop')
            df['bin'] = pd.cut(df['x'], bins=bins, include_lowest=True)
            grouped = df.groupby('bin', observed=False)['y'].agg(['count', 'sum'])
            grouped['bad'] = grouped['sum']
            grouped['good'] = grouped['count'] - grouped['sum']
            grouped = grouped.reset_index()
            intervals = grouped['bin'].tolist()
            cutoffs = [i.right for i in intervals]
            cutoffs = cutoffs[:-1]
            stats = grouped[['good', 'bad']].values
        else:
            grouped = df.groupby('x')['y'].agg(['count', 'sum'])
            grouped['bad'] = grouped['sum']
            grouped['good'] = grouped['count'] - grouped['sum']
            stats = grouped[['good', 'bad']].values
            cutoffs = grouped.index.tolist()[:-1]

        total_samples = len(df)
        min_samples = total_samples * self.min_bin_pct
        
        # Merge Logic
        while True:
            # 1. Check for small bins and merge them first
            # Calculate bin sizes
            # stats is [[good, bad], ...]
            # Bin i corresponds to interval (cutoffs[i-1], cutoffs[i]] (logic is roughly this)
            # Actually, len(stats) = len(cutoffs) + 1
            
            bin_sizes = stats.sum(axis=1)
            small_bin_indices = np.where(bin_sizes < min_samples)[0]
            
            if len(stats) <= 2: # Don't merge if only 2 bins left
                break
                
            merge_idx = -1
            
            # If there are small bins, prioritize merging the smallest one
            if len(small_bin_indices) > 0:
                # Find smallest bin
                min_size_idx = np.argmin(bin_sizes)
                # If smallest bin is >= min_samples, then no small bins (logic check)
                if bin_sizes[min_size_idx] >= min_samples:
                    # Proceed to Chi2 check
                    pass
                else:
                    # Merge smallest bin with neighbor having lower Chi2
                    idx = min_size_idx
                    if idx == 0:
                        merge_idx = 0 # Merge 0 and 1
                    elif idx == len(stats) - 1:
                        merge_idx = idx - 1 # Merge last and second last
                    else:
                        # Calculate Chi2 with left and right
                        chi2_left = self._calculate_chi2(stats[idx-1], stats[idx])
                        chi2_right = self._calculate_chi2(stats[idx], stats[idx+1])
                        if chi2_left < chi2_right:
                            merge_idx = idx - 1
                        else:
                            merge_idx = idx
            
            # If no small bins need merging, check max_bins
            if merge_idx == -1:
                if len(stats) <= max_bins:
                    # Satisfied max_bins, now enforce monotonicity if needed
                    # We do this as a post-check to avoid over-merging early?
                    # Or we do it iteratively.
                    
                    # If we are here, we have <= max_bins and no small bins.
                    # Now check trend.
                    break # We break here and do a final pass for monotonicity
                
                # Merge based on min Chi2
                chi2_list = []
                for i in range(len(stats) - 1):
                    chi2_list.append(self._calculate_chi2(stats[i], stats[i+1]))
                
                merge_idx = np.argmin(chi2_list)
            
            # Execute Merge
            stats[merge_idx] += stats[merge_idx+1]
            stats = np.delete(stats, merge_idx+1, axis=0)
            del cutoffs[merge_idx]
            
        # Post-processing: Enforce Monotonicity
        # If constraint is None, detect it automatically
        if constraint is None:
            constraint = self._detect_trend(x, y)
            
        if constraint in [1, -1]:
            stats, cutoffs = self._enforce_monotonicity(stats, cutoffs, constraint)
            
        return cutoffs

    def _calculate_chi2(self, bin1, bin2):
        obs = np.array([bin1, bin2])
        row_sums = obs.sum(axis=1)
        col_sums = obs.sum(axis=0)
        total = obs.sum()
        
        if 0 in row_sums or 0 in col_sums:
            return 0
            
        expected = np.outer(row_sums, col_sums) / total
        with np.errstate(divide='ignore', invalid='ignore'):
            chi2 = ((obs - expected)**2 / expected).sum()
        return chi2


    def _dt_binning(self, x, y, max_bins=10):
        # ... implementation using DecisionTreeClassifier ...
        df = pd.DataFrame({'x': x, 'y': y}).dropna()
        dt = DecisionTreeClassifier(max_leaf_nodes=max_bins, min_samples_leaf=0.05)
        dt.fit(df[['x']], df['y'])
        thresholds = sorted(dt.tree_.threshold[dt.tree_.threshold != -2])
        return thresholds

    def _quantile_binning(self, x, max_bins=10):
        _, bins = pd.qcut(x, max_bins, retbins=True, duplicates='drop')
        return bins[1:-1].tolist()

    def _calculate_woe(self, x, y, cutoffs):
        """
        Calculate WOE and IV for given cutoffs.
        """
        # bins = [-inf] + cutoffs + [inf]
        bins = [-np.inf] + cutoffs + [np.inf]
        
        df = pd.DataFrame({'x': x, 'y': y})
        df['bin'] = pd.cut(df['x'], bins=bins, include_lowest=True, labels=False)
        
        # Handle Missing
        # Map nan to -1
        df['bin'] = df['bin'].fillna(-1)
        
        total_good = (df['y'] == 0).sum()
        total_bad = (df['y'] == 1).sum()
        
        grouped = df.groupby('bin', observed=False)['y'].agg(['count', 'sum'])
        grouped['bad'] = grouped['sum']
        grouped['good'] = grouped['count'] - grouped['sum']
        
        woe_map = {}
        iv_total = 0
        bin_stats = []
        
        for bin_idx, row in grouped.iterrows():
            good = row['good']
            bad = row['bad']
            
            # Smooth to avoid inf
            if good == 0: good = 0.5
            if bad == 0: bad = 0.5
            
            dist_good = good / total_good
            dist_bad = bad / total_bad
            
            woe = np.log(dist_bad / dist_good)
            iv = (dist_bad - dist_good) * woe
            
            woe_map[bin_idx] = woe
            iv_total += iv
            
            bin_stats.append({
                'bin_index': bin_idx,
                'min': bins[int(bin_idx)] if bin_idx != -1 else np.nan,
                'max': bins[int(bin_idx)+1] if bin_idx != -1 else np.nan,
                'good': row['good'],
                'bad': row['bad'],
                'total': row['count'],
                'bad_rate': row['bad'] / row['count'],
                'woe': woe,
                'iv': iv
            })
            
        return woe_map, iv_total, bin_stats



import numpy as np

class Config:
    # Random Seed
    RANDOM_STATE = 42
    
    # Missing Values
    MISSING_VALUES = [float('nan'), None, np.nan]
    
    # Special Values that should be treated as separate bins (e.g., -999, -1)
    SPECIAL_VALUES = []
    
    # Binning Defaults
    MAX_BINS = 10
    MIN_BIN_PCT = 0.05
    
    # Selection Defaults
    IV_THRESHOLD = 0.02
    PSI_THRESHOLD = 0.1
    CORR_THRESHOLD = 0.7
    
    # Modeling Defaults
    TEST_SIZE = 0.3
    
import numpy as np




import pandas as pd
import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import log_loss
from .config import Config

class FeatureSelector:
    def __init__(self):
        self.selected_features = []
        self.dropped_features = {}
        
    def filter_by_iv(self, iv_df, threshold=0.02):
        """
        Filter features based on Information Value.
        iv_df: DataFrame with 'variable' and 'iv' columns.
        """
        selected = iv_df[iv_df['iv'] >= threshold]['variable'].tolist()
        dropped = iv_df[iv_df['iv'] < threshold]['variable'].tolist()
        
        self.dropped_features['iv_filter'] = dropped
        return selected
        
    def filter_by_corr(self, df, features=None, threshold=0.7):
        """
        Filter features based on correlation.
        Keeps the one with higher IV if iv_df provided (not implemented here for simplicity, just first one).
        """
        if features is None:
            features = df.columns.tolist()
            
        df_sub = df[features]
        corr_matrix = df_sub.corr().abs()
        upper = corr_matrix.where(np.triu(np.ones(corr_matrix.shape), k=1).astype(bool))
        
        to_drop = [column for column in upper.columns if any(upper[column] > threshold)]
        
        selected = [f for f in features if f not in to_drop]
        self.dropped_features['corr_filter'] = to_drop
        return selected

    def stepwise_selection(self, X, y, initial_features=None, method='forward', criterion='aic'):
        """
        Stepwise selection using Logistic Regression and AIC/BIC.
        """
        if initial_features is None:
            remaining = list(X.columns)
            selected = []
        else:
            remaining = [f for f in X.columns if f not in initial_features]
            selected = list(initial_features)
            
        current_score = np.inf
        best_new_score = np.inf
        
        # Initial score
        if selected:
            current_score = self._calculate_score(X[selected], y, criterion)
        else:
            # Null model score (intercept only)
            # We simulate intercept only by passing a column of ones if using fit_intercept=False
            # But sklearn LR adds intercept by default.
            # So passing empty X is tricky.
            # We calculate score for null model manually or using DummyClassifier
            # Or just start with inf and let the first feature be added.
            null_lr = LogisticRegression(penalty=None, solver='lbfgs')
            # Fake feature of 1s
            null_lr.fit(np.ones((len(y), 1)), y)
            probs = null_lr.predict_proba(np.ones((len(y), 1)))
            ll = log_loss(y, probs)
            k = 1 # intercept
            n = len(y)
            if criterion == 'aic':
                current_score = 2*k + 2*n*ll
            else:
                current_score = np.log(n)*k + 2*n*ll
        
        while remaining:
            scores_with_candidates = []
            
            for candidate in remaining:
                features_to_test = selected + [candidate]
                score = self._calculate_score(X[features_to_test], y, criterion)
                scores_with_candidates.append((score, candidate))
                
            scores_with_candidates.sort()
            best_new_score, best_candidate = scores_with_candidates[0]
            
            if best_new_score < current_score:
                remaining.remove(best_candidate)
                selected.append(best_candidate)
                current_score = best_new_score
                # print(f"Added {best_candidate}, score: {current_score}")
            else:
                break
                
        self.selected_features = selected
        return selected

    def _calculate_score(self, X, y, criterion='aic'):
        """
        Calculate AIC/BIC for Logistic Regression.
        """
        # Use simple model without penalty for stepwise to be faster and standard
        # However, if features are correlated, it might fail.
        # Adding a small L2 regularization helps convergence.
        lr = LogisticRegression(penalty='l2', C=1e9, solver='lbfgs', max_iter=1000)
        try:
            lr.fit(X, y)
            probs = lr.predict_proba(X)
            ll = log_loss(y, probs)
            
            k = X.shape[1] + 1 # +1 for intercept
            n = len(y)
            
            if criterion == 'aic':
                return 2*k + 2*n*ll
            elif criterion == 'bic':
                return np.log(n)*k + 2*n*ll
        except:
            return np.inf




import pandas as pd
import numpy as np
import joblib
from sklearn.model_selection import train_test_split

from .config import Config
from .eda import FeatureAnalyzer
from .binning import BinningProcess
from .selection import FeatureSelector
from .modeling import ModelTrainer
from .evaluation import ModelEvaluator
from .scorecard import ScorecardTransformer
from .reporting import ReportGenerator
from .plot_utils import plot_binning

class AutoScore:
    def __init__(self, random_state=42):
        self.random_state = random_state
        self.eda = FeatureAnalyzer()
        self.binning = None
        self.selector = FeatureSelector()
        self.trainer = ModelTrainer(random_state=random_state)
        self.evaluator = ModelEvaluator()
        self.sc_transformer = None
        self.reporter = ReportGenerator()
        
        # Intermediate results
        self.data_summary = None
        self.iv_df = None
        self.selected_features = []
        self.model_metrics = {}
        self.scorecard_df = None
        
    def fit(self, X, y, 
            test_size=0.3,
            date_col='create_time',
            oot_split_date=None,
            excluded_features=None,
            special_values=None,
            monotonicity_constraints=None,
            max_bins=10,
            min_bin_pct=0.05,
            iv_threshold=0.02,
            selection_method='stepwise',
            pdo=20, base_score=600, base_odds=50,
            custom_binning_rules=None,
            final_features=None):
        """
        Execute the full scorecard development pipeline.
        """
        # Preprocessing: Exclude features
        if excluded_features:
            X = X.drop(columns=[c for c in excluded_features if c in X.columns])
            
        # 1. Split Data (Train/Test/OOT)
        X_oot, y_oot = None, None
        
        if date_col in X.columns and oot_split_date:
            print(f"Splitting OOT data based on {date_col} > {oot_split_date}")
            mask_oot = X[date_col] > oot_split_date
            X_oot = X[mask_oot].copy()
            y_oot = y[mask_oot].copy()
            X = X[~mask_oot].copy()
            y = y[~mask_oot].copy()
            
            # Drop date column from modeling
            X_oot = X_oot.drop(columns=[date_col])
            X = X.drop(columns=[date_col])
        elif date_col in X.columns:
             X = X.drop(columns=[date_col])
        
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, test_size=test_size, random_state=self.random_state, stratify=y
        )
        
        # 2. EDA
        print("Running EDA...")
        self.data_summary = self.eda.analyze(X_train)
        
        # 3. Binning
        print("Running Binning...")
        self.binning = BinningProcess(max_bins=max_bins, min_bin_pct=min_bin_pct, special_values=special_values)
        
        if custom_binning_rules:
            print("Using custom binning rules...")
            self.binning.set_rules(custom_binning_rules)
            # Recalculate stats for the current training data using these rules
            self.binning.recalculate_stats(X_train, y_train)
        else:
            self.binning.fit(X_train, y_train, monotonicity_constraints=monotonicity_constraints)
        
        self.iv_df = self.binning.iv_df
        
        # Transform to WOE
        X_train_woe = self.binning.transform(X_train)
        X_test_woe = self.binning.transform(X_test)
        X_oot_woe = self.binning.transform(X_oot) if X_oot is not None else None
        
        # 4. Feature Selection
        print("Running Feature Selection...")
        if final_features:
             print(f"Using provided final features ({len(final_features)})...")
             self.selected_features = final_features
             # Filter iv_df for report consistency if needed, but not strictly required for flow
        else:
            # IV Filter
            features_iv = self.selector.filter_by_iv(self.iv_df, threshold=iv_threshold)
            print(f"Features after IV filter: {len(features_iv)}")
            
            # Stepwise
            if selection_method == 'stepwise':
                self.selected_features = self.selector.stepwise_selection(
                    X_train_woe[features_iv], y_train, method='forward'
                )
            else:
                self.selected_features = features_iv
            
        print(f"Final selected features: {len(self.selected_features)}")
        print(self.selected_features)
        
        # 5. Model Training
        print("Training Model...")
        self.trainer.train(X_train_woe[self.selected_features], y_train)
        
        # 6. Evaluation
        print("Evaluating Model...")
        self.evaluate_performance(X_train_woe, y_train, X_test_woe, y_test, X_oot_woe, y_oot)
        
        # 7. Scorecard Generation
        print("Generating Scorecard...")
        self.sc_transformer = ScorecardTransformer(pdo=pdo, base_score=base_score, base_odds=base_odds)
        self.scorecard_df = self.sc_transformer.fit(self.trainer.model, self.binning, self.selected_features)
        
        return self

    def evaluate_performance(self, X_train_woe, y_train, X_test_woe, y_test, X_oot_woe=None, y_oot=None):
        """
        Evaluate model on Train/Test/OOT.
        """
        # Train metrics
        pred_train = self.trainer.predict_proba(X_train_woe[self.selected_features])
        metrics_train = self.evaluator.evaluate(y_train, pred_train)
        
        # Test metrics
        pred_test = self.trainer.predict_proba(X_test_woe[self.selected_features])
        metrics_test = self.evaluator.evaluate(y_test, pred_test)
        
        # PSI (Train vs Test)
        psi_test = self.evaluator.calculate_psi(pred_train, pred_test)
        
        self.model_metrics = {
            'train': metrics_train,
            'test': metrics_test,
            'psi_train_test': psi_test
        }
        
        if X_oot_woe is not None:
            pred_oot = self.trainer.predict_proba(X_oot_woe[self.selected_features])
            metrics_oot = self.evaluator.evaluate(y_oot, pred_oot)
            psi_oot = self.evaluator.calculate_psi(pred_train, pred_oot)
            self.model_metrics['oot'] = metrics_oot
            self.model_metrics['psi_train_oot'] = psi_oot

    def update_model_weights(self, coef_dict, intercept=None):
        """
        Manually update model weights and re-evaluate.
        coef_dict: {feature: weight}
        """
        if self.trainer.model is None:
            print("Model not trained yet.")
            return
            
        # Get estimator
        estimator = self.trainer.model.best_estimator_ if hasattr(self.trainer.model, 'best_estimator_') else self.trainer.model
        
        # Update coefficients
        # Assuming coef_dict contains all selected features
        new_coefs = []
        for feat in self.selected_features:
            if feat in coef_dict:
                new_coefs.append(coef_dict[feat])
            else:
                # Keep original? Or error?
                # Let's try to find index
                idx = self.selected_features.index(feat)
                new_coefs.append(estimator.coef_[0][idx])
                
        estimator.coef_ = np.array([new_coefs])
        
        if intercept is not None:
            estimator.intercept_ = np.array([intercept])
            
        print("Model weights updated.")
        # Note: Should re-run evaluation and scorecard generation ideally
        # But user needs to provide data for evaluation. 
        # So usually this is done before final save.
        
    def adjust_rules(self, new_rules, X_train=None, y_train=None):
        """
        Manually adjust binning rules and visualize.
        """
        if self.binning is None:
            print("Binning process not initialized. Run fit() first or provide custom rules.")
            return
            
        print("Adjusting rules...")
        # Update specific rules
        for col, rule in new_rules.items():
            if col in self.binning.rules:
                # Assuming user provides 'cutoffs' or 'categories' directly
                if 'cutoffs' in rule:
                    self.binning.rules[col]['cutoffs'] = rule['cutoffs']
                if 'categories' in rule:
                    self.binning.rules[col]['categories'] = rule['categories']
            else:
                print(f"Warning: {col} not found in existing rules.")
                
        # Recalculate stats if data provided (highly recommended)
        if X_train is not None and y_train is not None:
            self.binning.recalculate_stats(X_train, y_train)
            self.iv_df = self.binning.iv_df
            
    def visualize_binning(self, feature_name, filename=None, show=True):
        """
        Visualize binning for a feature.
        """
        if self.binning is None or feature_name not in self.binning.rules:
            print(f"No binning info for {feature_name}")
            return
            
        stats = self.binning.rules[feature_name]['stats']
        plot_binning(stats, feature_name, show=show, filename=filename)

    def save_model(self, filepath):
        """Save the entire AutoScore object."""
        joblib.dump(self, filepath)
        
    def load_model(self, filepath):
        """Load AutoScore object."""
        return joblib.load(filepath)
        
    def predict_proba(self, X):
        X_woe = self.binning.transform(X)
        return self.trainer.predict_proba(X_woe[self.selected_features])
        
    def predict_score(self, X):
        X_woe = self.binning.transform(X)
        return self.sc_transformer.transform(X_woe, self.selected_features, self.trainer.model)
        
    def create_report(self, filename='scorecard_report.xlsx'):
        """
        Export all results to Excel.
        """
        # Prepare DataFrames
        metrics_list = [
            {'Metric': 'Train AUC', 'Value': self.model_metrics['train']['auc']},
            {'Metric': 'Train KS', 'Value': self.model_metrics['train']['ks']},
            {'Metric': 'Test AUC', 'Value': self.model_metrics['test']['auc']},
            {'Metric': 'Test KS', 'Value': self.model_metrics['test']['ks']},
            {'Metric': 'PSI (Train vs Test)', 'Value': self.model_metrics['psi_train_test']}
        ]
        
        if 'oot' in self.model_metrics:
            metrics_list.extend([
                {'Metric': 'OOT AUC', 'Value': self.model_metrics['oot']['auc']},
                {'Metric': 'OOT KS', 'Value': self.model_metrics['oot']['ks']},
                {'Metric': 'PSI (Train vs OOT)', 'Value': self.model_metrics['psi_train_oot']}
            ])
            
        metrics_df = pd.DataFrame(metrics_list)
        
        # Model Coefficients
        model = self.trainer.model.best_estimator_ if hasattr(self.trainer.model, 'best_estimator_') else self.trainer.model
        coef_df = pd.DataFrame({
            'Feature': ['Intercept'] + self.selected_features,
            'Coefficient': [model.intercept_[0]] + list(model.coef_[0])
        })
        
        data_dict = {
            'Summary': self.data_summary,
            'IV Analysis': self.iv_df,
            'Selected Features': pd.DataFrame(self.selected_features, columns=['Feature']),
            'Model Coefficients': coef_df,
            'Model Performance': metrics_df,
            'Scorecard': self.scorecard_df
        }
        
        return self.reporter.generate(filename, data_dict)




import pandas as pd
import numpy as np
from sklearn.metrics import roc_auc_score, roc_curve

class ModelEvaluator:
    def evaluate(self, y_true, y_pred_proba):
        """
        Calculate AUC, KS.
        """
        auc = roc_auc_score(y_true, y_pred_proba)
        
        fpr, tpr, thresholds = roc_curve(y_true, y_pred_proba)
        ks = max(tpr - fpr)
        
        return {
            'auc': auc,
            'ks': ks
        }
        
    def calculate_psi(self, expected, actual, buckets=10):
        """
        Calculate PSI between two distributions.
        expected: training scores (or reference)
        actual: test scores (or current)
        """
        def scale_range(input, min, max):
            input += -(np.min(input))
            input /= np.max(input) / (max - min)
            input += min
            return input

        breakpoints = np.arange(0, buckets + 1) / (buckets) * 100
        breakpoints = np.percentile(expected, breakpoints)
        
        # Avoid duplicate breakpoints
        breakpoints = sorted(list(set(breakpoints)))
        
        # Calculate counts
        expected_percents = np.histogram(expected, breakpoints)[0] / len(expected)
        actual_percents = np.histogram(actual, breakpoints)[0] / len(actual)
        
        def sub_psi(e_perc, a_perc):
            if a_perc == 0: a_perc = 0.0001
            if e_perc == 0: e_perc = 0.0001
            
            value = (e_perc - a_perc) * np.log(e_perc / a_perc)
            return value
            
        psi_value = np.sum([sub_psi(expected_percents[i], actual_percents[i]) for i in range(len(expected_percents))])
        
        return psi_value




import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

def plot_binning(bin_stats, feature_name, show=True, filename=None):
    """
    Visualize binning results with dual-axis chart (Count vs Bad Rate).
    bin_stats: list of dicts from BinningProcess
    """
    if not bin_stats:
        print(f"No binning stats available for {feature_name}")
        return

    df_stats = pd.DataFrame(bin_stats)
    
    # Create label
    df_stats['label'] = df_stats.apply(lambda x: f"[{x['min']}, {x['max']})", axis=1)
    
    # Plotting
    fig, ax1 = plt.subplots(figsize=(10, 6))
    
    # Bar chart for Total Count
    ax1.bar(df_stats.index, df_stats['total'], color='skyblue', alpha=0.6, label='Count')
    ax1.set_ylabel('Sample Count', color='blue')
    ax1.tick_params(axis='y', labelcolor='blue')
    ax1.set_xticks(df_stats.index)
    ax1.set_xticklabels(df_stats['label'], rotation=45, ha='right')
    ax1.set_xlabel('Bin')
    
    # Line chart for Bad Rate
    ax2 = ax1.twinx()
    ax2.plot(df_stats.index, df_stats['bad_rate'], color='red', marker='o', linestyle='-', linewidth=2, label='Bad Rate')
    ax2.set_ylabel('Bad Rate', color='red')
    ax2.tick_params(axis='y', labelcolor='red')
    
    # Add IV info
    total_iv = df_stats['iv'].sum()
    plt.title(f"Binning: {feature_name} (IV: {total_iv:.4f})")
    
    # Layout
    plt.tight_layout()
    
    if filename:
        plt.savefig(filename)
        
    if show:
        plt.show()
    else:
        plt.close()
